@@ -8,12 +8,10 @@
 # copied, modified, or distributed except according to those terms.
 import logging
 import os.path
-from base64 import b32decode
 from hashlib import sha256
-from typing import BinaryIO, List, Optional, Type, TypeVar
+from typing import BinaryIO, Callable, List, Optional, Type, TypeVar
 
 import click
-import fido2
 from cryptography import x509
 from cryptography.hazmat.primitives import serialization
 from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
@@ -21,25 +19,22 @@ from ecdsa import NIST256p, SigningKey
 
 from pynitrokey.cli.exceptions import CliException
 from pynitrokey.helpers import (
-    AskUser,
     DownloadProgressBar,
     Retries,
     local_print,
-    prompt,
     require_windows_admin,
 )
 from pynitrokey.nk3 import list as list_nk3
 from pynitrokey.nk3 import open as open_nk3
+from pynitrokey.nk3.admin_app import AdminApp
 from pynitrokey.nk3.base import Nitrokey3Base
 from pynitrokey.nk3.bootloader import (
+    FirmwareContainer,
     Nitrokey3Bootloader,
-    Variant,
-    detect_variant,
     parse_firmware_image,
 )
 from pynitrokey.nk3.device import BootMode, Nitrokey3Device
 from pynitrokey.nk3.exceptions import TimeoutException
-from pynitrokey.nk3.otp_app import STRING_TO_KIND, Algorithm, OTPApp
 from pynitrokey.nk3.provisioner_app import ProvisionerApp
 from pynitrokey.nk3.updates import REPOSITORY, get_firmware_update
 from pynitrokey.updates import OverwriteError
@@ -48,17 +43,6 @@ T = TypeVar("T", bound=Nitrokey3Base)
 
 logger = logging.getLogger(__name__)
 
-VARIANT_CHOICE = click.Choice(
-    [variant.value for variant in Variant],
-    case_sensitive=False,
-)
-
-
-def variant_callback(
-    _ctx: object, _param: object, value: Optional[str]
-) -> Optional[Variant]:
-    return Variant.from_str(value) if value else None
-
 
 class Context:
     def __init__(self, path: Optional[str]) -> None:
@@ -97,25 +81,45 @@ class Context:
         ]
         return self._select_unique("Nitrokey 3", devices)
 
-    def _await(self, name: str, ty: Type[T]) -> T:
-        for t in Retries(10):
+    def _await(
+        self,
+        name: str,
+        ty: Type[T],
+        retries: int,
+        callback: Optional[Callable[[int, int], None]] = None,
+    ) -> T:
+        for t in Retries(retries):
             logger.debug(f"Searching {name} device ({t})")
             devices = [device for device in self.list() if isinstance(device, ty)]
             if len(devices) == 0:
+                if callback:
+                    callback(int((t.i / retries) * 100), 100)
                 logger.debug(f"No {name} device found, continuing")
                 continue
             if len(devices) > 1:
                 raise CliException(f"Multiple {name} devices found")
+            if callback:
+                callback(100, 100)
             return devices[0]
 
         raise CliException(f"No {name} device found")
 
-    def await_device(self) -> Nitrokey3Device:
-        return self._await("Nitrokey 3", Nitrokey3Device)
+    def await_device(
+        self,
+        retries: Optional[int] = 30,
+        callback: Optional[Callable[[int, int], None]] = None,
+    ) -> Nitrokey3Device:
+        assert isinstance(retries, int)
+        return self._await("Nitrokey 3", Nitrokey3Device, retries, callback)
 
-    def await_bootloader(self) -> Nitrokey3Bootloader:
+    def await_bootloader(
+        self,
+        retries: Optional[int] = 30,
+        callback: Optional[Callable[[int, int], None]] = None,
+    ) -> Nitrokey3Bootloader:
+        assert isinstance(retries, int)
         # mypy does not allow abstract types here, but this is still valid
-        return self._await("Nitrokey 3 bootloader", Nitrokey3Bootloader)  # type: ignore
+        return self._await("Nitrokey 3 bootloader", Nitrokey3Bootloader, retries, callback)  # type: ignore
 
 
 @click.group()
@@ -309,13 +313,6 @@ def test(
 
 @nk3.command()
 @click.argument("path", default=".")
-@click.option(
-    "--variant",
-    type=VARIANT_CHOICE,
-    callback=variant_callback,
-    required=True,
-    help="The variant to fetch the update for",
-)
 @click.option(
     "-f",
     "--force",
@@ -324,9 +321,7 @@ def test(
     help="Overwrite the firmware image if it already exists",
 )
 @click.option("--version", help="Download this version instead of the latest one")
-def fetch_update(
-    path: str, force: bool, variant: Variant, version: Optional[str]
-) -> None:
+def fetch_update(path: str, force: bool, version: Optional[str]) -> None:
     """
     Fetches a firmware update for the Nitrokey 3 and stores it at the given path.
 
@@ -340,7 +335,7 @@ def fetch_update(
     """
     try:
         release = REPOSITORY.get_release_or_latest(version)
-        update = get_firmware_update(release, variant)
+        update = get_firmware_update(release)
     except Exception as e:
         if version:
             raise CliException(f"Failed to find firmware update {version}", e)
@@ -372,47 +367,47 @@ def fetch_update(
 
 @nk3.command()
 @click.argument("image")
-@click.option(
-    "--variant",
-    type=VARIANT_CHOICE,
-    callback=variant_callback,
-    help="The variant of the given firmage image",
-)
-def validate_update(image: str, variant: Optional[Variant]) -> None:
+def validate_update(image: str) -> None:
     """
-    Validates the given firmware image and prints the firmware version and the signer.
-
-    If the name of the firmware image name is changed so that the device variant can no longer be
-    detected from the filename, it has to be set explictly with --variant.
+    Validates the given firmware image and prints the firmware version and the signer for all
+    available variants.
     """
-    if not variant:
-        variant = detect_variant(image)
-    if not variant:
-        variant = Variant.from_str(
-            prompt("Firmware image variant", type=VARIANT_CHOICE)
-        )
+    container = FirmwareContainer.parse(image)
+    print(f"version:      {container.version}")
+    if container.pynitrokey:
+        print(f"pynitrokey:   >= {container.pynitrokey}")
 
-    with open(image, "rb") as f:
-        data = f.read()
+    for variant in container.images:
+        data = container.images[variant]
+        try:
+            metadata = parse_firmware_image(variant, data)
+        except Exception as e:
+            raise CliException("Failed to parse and validate firmware image", e)
 
-    try:
-        metadata = parse_firmware_image(variant, data)
-    except Exception as e:
-        raise CliException("Failed to parse and validate firmware image", e)
+        signed_by = metadata.signed_by or "unsigned"
 
-    signed_by = metadata.signed_by or "unsigned"
+        print(f"variant:      {variant.value}")
+        print(f"  version:    {metadata.version}")
+        print(f"  signed by:  {signed_by}")
 
-    print(f"version:    {metadata.version}")
-    print(f"signed by:  {signed_by}")
+        if container.version != metadata.version:
+            raise CliException(
+                f"The firmware image for the {variant} variant and the release {version} has an "
+                f"unexpected product version ({metadata.version})."
+            )
 
 
 @nk3.command()
 @click.argument("image", required=False)
 @click.option(
-    "--variant",
-    type=VARIANT_CHOICE,
-    callback=variant_callback,
-    help="The variant of the given firmage image",
+    "--version",
+    help="Set the firmware version to update to (default: latest stable)",
+)
+@click.option(
+    "--ignore-pynitrokey-version",
+    default=False,
+    is_flag=True,
+    help="Allow updates with an outdated pynitrokey version (dangerous)",
 )
 @click.option(
     "--experimental",
@@ -423,7 +418,11 @@ def validate_update(image: str, variant: Optional[Variant]) -> None:
 )
 @click.pass_obj
 def update(
-    ctx: Context, image: Optional[str], variant: Optional[Variant], experimental: bool
+    ctx: Context,
+    image: Optional[str],
+    version: Optional[str],
+    ignore_pynitrokey_version: bool,
+    experimental: bool,
 ) -> None:
     """
     Update the firmware of the device using the given image.
@@ -433,9 +432,8 @@ def update(
     not be removed during the update.  Also, additional Nitrokey 3 devices may not be connected
     during the update.
 
-    If no firmware image is given, the latest firmware release is downloaded automatically.  If a
-    firmware image is given and its name is changed so that the device variant can no longer be
-    detected from the filename, it has to be set explictly with --variant.
+    If no firmware image is given, the latest firmware release is downloaded automatically.  If
+    the --version option is set, the given version is downloaded instead.
 
     If the connected Nitrokey 3 device is in firmware mode, the user is prompted to touch the
     device’s button to confirm rebooting to bootloader mode.
@@ -446,18 +444,29 @@ def update(
 
     from .update import update as exec_update
 
-    update_version = exec_update(ctx, image, variant)
+    exec_update(ctx, image, version, ignore_pynitrokey_version)
 
-    local_print("")
-    with ctx.await_device() as device:
-        version = device.version()
-        if version == update_version:
-            local_print(f"Successfully updated the firmware to version {version}.")
-        else:
-            raise CliException(
-                f"The firmware update to {update_version} was successful, but the firmware "
-                f"is still reporting version {version}."
-            )
+
+@nk3.command()
+@click.pass_obj
+def status(ctx: Context) -> None:
+    """Query the device status."""
+    with ctx.connect_device() as device:
+        uuid = device.uuid()
+        if uuid is not None:
+            local_print(f"UUID:               {uuid}")
+
+        admin = AdminApp(device)
+        version = admin.version()
+        local_print(f"Firmware version:   {version}")
+
+        status = admin.status()
+        if status.init_status is not None:
+            local_print(f"Init status:        {status.init_status}")
+        if status.ifs_blocks is not None:
+            local_print(f"Free blocks (int):  {status.ifs_blocks}")
+        if status.efs_blocks is not None:
+            local_print(f"Free blocks (ext):  {status.efs_blocks}")
 
 
 @nk3.command()
@@ -477,326 +486,6 @@ def wink(ctx: Context) -> None:
         device.wink()
 
 
-@nk3.group()
-@click.pass_context
-def otp(ctx: click.Context) -> None:
-    """Manage OTP secrets on the device.
-    Use NITROPY_OTP_PASSWORD to pass password for the scripted execution."""
-    pass
-
-
-@otp.command()
-@click.pass_obj
-@click.argument(
-    "name",
-    type=click.STRING,
-)
-@click.argument(
-    "secret",
-    type=click.STRING,
-    # help="The shared secret string (by default in base32)",
-)
-@click.option(
-    "--digits_str",
-    "digits_str",
-    type=click.Choice(["6", "8"]),
-    help="Digits count",
-    default="6",
-)
-@click.option(
-    "--kind",
-    "kind",
-    type=click.Choice(choices=STRING_TO_KIND.keys(), case_sensitive=False),  # type: ignore[arg-type]
-    help="OTP mechanism to use. Case insensitive.",
-    default="TOTP",
-)
-@click.option(
-    "--hash",
-    "hash",
-    type=click.Choice(["SHA1", "SHA256"]),
-    help="Hash algorithm to use",
-    default="SHA1",
-)
-@click.option(
-    "--counter_start",
-    "counter_start",
-    type=click.INT,
-    help="Starting value for the counter (HOTP only)",
-    default=0,
-)
-@click.option(
-    "--touch_button",
-    "touch_button",
-    type=click.BOOL,
-    help="This credential requires button press before use",
-    is_flag=True,
-)
-@click.option(
-    "--experimental",
-    default=False,
-    is_flag=True,
-    help="Allow to execute experimental features",
-)
-def register(
-    ctx: Context,
-    name: str,
-    secret: str,
-    digits_str: str,
-    kind: str,
-    hash: str,
-    counter_start: int,
-    touch_button: bool,
-    experimental: bool,
-) -> None:
-    """Register OTP credential.
-
-    Write SECRET under the NAME.
-    SECRET should be encoded in base32 format.
-    Experimental.
-    """
-    check_experimental_flag(experimental)
-
-    digits = int(digits_str)
-    secret_bytes = b32decode(secret)
-    otp_kind = STRING_TO_KIND[kind.upper()]
-    hash_algorithm = Algorithm.Sha1 if hash == "SHA1" else Algorithm.Sha256
-    with ctx.connect_device() as device:
-        app = OTPApp(device)
-        authenticate_if_needed(app)
-        app.register(
-            name.encode(),
-            secret_bytes,
-            digits,
-            kind=otp_kind,
-            algo=hash_algorithm,
-            initial_counter_value=counter_start,
-            touch_button_required=touch_button,
-        )
-
-
-def check_experimental_flag(experimental: bool) -> None:
-    """Helper function to show common warning for the experimental features"""
-    if not experimental:
-        local_print(" ")
-        local_print(
-            "This feature is experimental, which means it was not tested thoroughly.\n"
-            "Note: data stored with it can be lost in the next firmware update.\n"
-            "Please pass --experimental switch to force running it anyway."
-        )
-        local_print(" ")
-        raise click.Abort()
-
-
-@otp.command()
-@click.pass_obj
-@click.option(
-    "--hex",
-    "hex",
-    type=click.BOOL,
-    help="Use hex representation",
-    default=False,
-    is_flag=True,
-)
-def show(ctx: Context, hex: bool) -> None:
-    """List registered OTP credentials."""
-    with ctx.connect_device() as device:
-        app = OTPApp(device)
-        authenticate_if_needed(app)
-        for e in app.list():
-            local_print(e.hex() if hex else e)
-
-
-@otp.command()
-@click.pass_obj
-@click.argument(
-    "name",
-    type=click.STRING,
-)
-def remove(ctx: Context, name: str) -> None:
-    """Remove OTP credential."""
-    with ctx.connect_device() as device:
-        app = OTPApp(device)
-        authenticate_if_needed(app)
-        app.delete(name.encode())
-
-
-@otp.command()
-@click.pass_obj
-@click.option(
-    "--force",
-    is_flag=True,
-    help="Do not ask for confirmation",
-)
-def reset(ctx: Context, force: bool) -> None:
-    """Remove all OTP credentials from the device."""
-    confirmed = force or click.confirm("Do you want to continue?")
-    if not confirmed:
-        local_print("Operation cancelled")
-        click.Abort()
-    with ctx.connect_device() as device:
-        app = OTPApp(device)
-        app.reset()
-        local_print("Operation executed")
-
-
-@otp.command()
-@click.pass_obj
-@click.argument(
-    "name",
-    type=click.STRING,
-)
-@click.option(
-    "--timestamp",
-    "timestamp",
-    type=click.INT,
-    help="The timestamp to use instead of the local time (TOTP only)",
-    default=0,
-)
-@click.option(
-    "--period",
-    "period",
-    type=click.INT,
-    help="The period to use in seconds (TOTP only)",
-    default=30,
-)
-@click.option(
-    "--experimental",
-    default=False,
-    is_flag=True,
-    help="Allow to execute experimental features",
-)
-def get(
-    ctx: Context, name: str, timestamp: int, period: int, experimental: bool
-) -> None:
-    """Generate OTP code from registered credential.
-    Experimental."""
-    # TODO: for TOTP get the time from a timeserver via NTP, instead of the local clock
-    check_experimental_flag(experimental)
-
-    from datetime import datetime
-
-    now = datetime.now()
-    timestamp = timestamp if timestamp else int(datetime.timestamp(now))
-    with ctx.connect_device() as device:
-        try:
-            app = OTPApp(device)
-            authenticate_if_needed(app)
-            code = app.calculate(name.encode(), timestamp // period)
-            local_print(
-                f"Timestamp: {datetime.isoformat(now, timespec='seconds')} ({timestamp}), period: {period}"
-            )
-            local_print(code.decode())
-        except fido2.ctap.CtapError as e:
-            local_print(
-                f"Device returns error: {e}. This credential id might not be registered."
-            )
-
-
-@otp.command()
-@click.pass_obj
-@click.argument(
-    "name",
-    type=click.STRING,
-)
-@click.option(
-    "--code",
-    "code",
-    type=click.INT,
-    help="The code to verify",
-    default=0,
-)
-@click.option(
-    "--experimental",
-    default=False,
-    is_flag=True,
-    help="Allow to execute experimental features",
-)
-def verify(ctx: Context, name: str, code: int, experimental: bool) -> None:
-    """Proceed with the incoming OTP code verification (aka reverse HOTP).
-    Does not need authentication by design.
-    Experimental."""
-    check_experimental_flag(experimental)
-
-    with ctx.connect_device() as device:
-        app = OTPApp(device)
-        try:
-            app.verify_code(name.encode(), code)
-        except fido2.ctap.CtapError as e:
-            local_print(
-                f"Device returns error: {e}. This credential id might not be registered, or the provided HOTP code has not passed verification."
-            )
-
-
-def ask_for_passphrase_if_needed(app: OTPApp) -> Optional[str]:
-    passphrase = None
-    if app.authentication_required():
-        passphrase = AskUser(
-            "Current Password", envvar="NITROPY_OTP_PASSWORD", hide_input=True
-        ).ask()
-    return passphrase
-
-
-def authenticate_if_needed(app: OTPApp) -> None:
-    try:
-        passphrase = ask_for_passphrase_if_needed(app)
-        if passphrase is not None:
-            app.validate(passphrase)
-    except Exception as e:
-        local_print(f'Authentication failed with error: "{e}"')
-        raise click.Abort()
-
-
-@otp.command()
-@click.pass_obj
-@click.option(
-    "--experimental",
-    default=False,
-    is_flag=True,
-    help="Allow to execute experimental features",
-)
-@click.password_option()
-def set_password(ctx: Context, password: str, experimental: bool) -> None:
-    """Set the passphrase used to authenticate to other commands.
-    Experimental."""
-    check_experimental_flag(experimental)
-
-    with ctx.connect_device() as device:
-        try:
-            app = OTPApp(device)
-            authenticate_if_needed(app)
-            app.set_code(password)
-            local_print("Password set")
-        except fido2.ctap.CtapError as e:
-            local_print(
-                f"Device returns error: {e}. This passphrase might be invalid or is set already."
-            )
-
-
-@otp.command()
-@click.pass_obj
-@click.option(
-    "--experimental",
-    default=False,
-    is_flag=True,
-    help="Allow to execute experimental features",
-)
-def clear_password(ctx: Context, experimental: bool) -> None:
-    """Clear the passphrase used to authenticate to other commands.
-    Experimental."""
-    check_experimental_flag(experimental)
-
-    with ctx.connect_device() as device:
-        try:
-            app = OTPApp(device)
-            authenticate_if_needed(app)
-            app.clear_code()
-            local_print("Password cleared")
-        except fido2.ctap.CtapError as e:
-            local_print(
-                f"Device returns error: {e}. This passphrase might be invalid or is set already."
-            )
-
-
 @nk3.group(hidden=True)
 def provision() -> None:
     """
@@ -886,3 +575,8 @@ def provision_fido2(ctx: Context, key_file: BinaryIO, cert_file: BinaryIO) -> No
         provisioner = ProvisionerApp(device)
         provisioner.write_file(b"fido/x5c/00", cert)
         provisioner.write_file(b"fido/sec/00", key)
+
+
+# This import has to be added here to avoid circular dependency
+# Import "secrets" subcommand from the secrets module
+from . import secrets  # noqa: F401,E402
